# Third Party
import logging

import tomlkit
from resc_backend.constants import RULE_TAG_SCAN_AS_DIR

from vcs_scanner.helpers.gitleaks_types import GitLeaksConfigToml, RuleToml

logger = logging.getLogger(__name__)


class RuleFileProvider:
    def __init__(self, toml_rule_file_path: str, init: bool = False):
        self.base_rule_file_path: str = toml_rule_file_path
        self.scan_as_repo_rule_file_path: str | None = toml_rule_file_path if init else None
        self.scan_as_dir_rule_file_path: str | None = toml_rule_file_path if init else None

    def init(self, destination_rule_as_repo: str, destination_rule_as_dir: str) -> None:
        if destination_rule_as_repo == self.base_rule_file_path and destination_rule_as_dir == self.base_rule_file_path:
            self.scan_as_repo_rule_file_path = destination_rule_as_repo
            self.scan_as_dir_rule_file_path = destination_rule_as_dir
            return None

        if self.scan_as_dir_rule_file_path is not None or self.scan_as_repo_rule_file_path is not None:
            return None

        toml_dict: GitLeaksConfigToml = {}
        rules_as_repo: list[RuleToml] = []
        rules_as_dir: list[RuleToml] = []

        # read toml
        with open(self.base_rule_file_path, encoding="utf-8") as toml_rule_file:
            toml_rule_dictionary = tomlkit.load(toml_rule_file)
            toml_dict = toml_rule_dictionary.unwrap()
            for rule in toml_dict.get("rules", []):
                (rules_as_dir if RULE_TAG_SCAN_AS_DIR in rule.get("tags", []) else rules_as_repo).append(rule)

        if len(rules_as_repo) > 0:
            if self._create_rule_file(toml_dict, rules_as_repo, destination_rule_as_repo):
                self.scan_as_repo_rule_file_path = destination_rule_as_repo

        if len(rules_as_dir) > 0:
            if self._create_rule_file(toml_dict, rules_as_dir, destination_rule_as_dir):
                self.scan_as_dir_rule_file_path = destination_rule_as_dir

    def _create_rule_file(self, toml_dict: GitLeaksConfigToml, rules: list[RuleToml], destination: str) -> bool:
        """
        Create a rule file given toml dictionary

        Args:
            toml_dict (dict): _description_
            rules (list): _description_
            destination (str): _description_

        Returns:
            bool: True on successfully created file.
        """

        new_toml_dict: GitLeaksConfigToml = {
            "title": toml_dict["title"],
            "version": toml_dict["version"],
            "rules": rules,
            "allowlist": toml_dict.get("allowlist", []),
        }

        try:
            with open(destination, "w") as f:
                tomlkit.dump(new_toml_dict, f)
        except OSError as err:
            logger.error(f"could not write in {destination}: {err}")
        except Exception as err:
            logger.error(f"Unexpected {err=}, {type(err)=}")
            return False

        return True



LOGGING_FILE = "logging.ini"
LOG_FILE_PATH_CLI = "cli_secret_scanner.log"
LOG_FILE_PATH = "secret_scanner.log"

LEAKS_FOUND_EXIT_CODE = 42
NO_LEAKS_FOUND_EXIT_CODE = 0

# These values are used as vcs names whenever a CLI scan is done
CLI_VCS_LOCAL_SCAN = "LOCAL_SCAN"
CLI_VCS_BITBUCKET = "CLI_BITBUCKET"
CLI_VCS_AZURE = "CLI_AZURE"



# pylint: disable=bad-option-value,C0413
# Standard Library
import logging
import os

os.environ["GIT_PYTHON_REFRESH"] = "quiet"

# Third Party
from git import Commit, Repo  # noqa: E402

logger = logging.getLogger(__name__)


def clone_repository(
    repository_url: str,
    repo_clone_path: str,
    username: str = "",
    personal_access_token: str = "",
) -> Commit:
    """
        Clones the given repository
    :param repository_url:
        Repository url to clone
    :param repo_clone_path:
        Path where to clone the repository
    :param username:
        Username to clone the repository, only needed if the repository is private
    :param personal_access_token:
        Personal access token|password to clone the repository, only needed if the repository is private
    """
    url = repository_url.replace("https://", "")
    repo_clone_url = f"https://{username}:{personal_access_token}@{url}"
    repo = Repo.clone_from(repo_clone_url, repo_clone_path)
    logger.debug(f"Repository {repository_url} cloned successfully")
    return repo.head.commit


def read_repo_from_local(path_to_dir: str) -> str:
    """Given a path returns the remote address of the repository

    Args:
        path_to_dir (str): Path to the repo (.git must be in that directory)

    Returns:
        str: Url of the repository
    """
    repo = Repo(path_to_dir)
    logger.debug(repo.remotes[0].url)
    return repo.remotes[0].url



# Standard Library
import abc

# Third Party
from resc_backend.resc_web_service.schema.finding import FindingCreate
from resc_backend.resc_web_service.schema.repository import Repository, RepositoryBase
from resc_backend.resc_web_service.schema.scan import Scan, ScanRead
from resc_backend.resc_web_service.schema.scan_type import ScanType
from resc_backend.resc_web_service.schema.vcs_instance import VCSInstanceRead

# First Party
from vcs_scanner.model import VCSInstanceRuntime


class OutputModule(metaclass=abc.ABCMeta):
    @abc.abstractmethod
    def load_rules(self, toml_rule_file_path: str) -> None:
        raise NotImplementedError

    @abc.abstractmethod
    def write_vcs_instance(self, vcs_instance_runtime: VCSInstanceRuntime) -> VCSInstanceRead | None:
        raise NotImplementedError

    @abc.abstractmethod
    def write_repository(self, repository: Repository) -> RepositoryBase | None:
        raise NotImplementedError

    @abc.abstractmethod
    def write_findings(
        self,
        scan_id: int,
        repository_id: int,
        scan_findings: list[FindingCreate],
        repository_name: str = "",
    ) -> None:
        raise NotImplementedError

    @abc.abstractmethod
    def write_scan(
        self,
        scan_type_to_run: ScanType,
        last_scanned_commit: str,
        scan_timestamp: str,
        repository: Repository,
        rule_pack: str,
    ) -> Scan:
        raise NotImplementedError

    @abc.abstractmethod
    def get_last_scan_for_repository(self, repository: Repository) -> ScanRead | None:
        raise NotImplementedError



# Third Party
from resc_backend.resc_web_service.schema.finding import FindingCreate


def should_process_finding(
    finding: FindingCreate,
    rule_tags: dict = None,
    include_tags: list[str] = None,
    ignore_tags: list[str] = None,
) -> bool:
    """
        Determine the action to take for the finding, based on the rule tags
    :param finding:
        FindingCreate instance of the finding
    :param rule_tags:
        Dictionary containing all the rules and there respective tags
    :param include_tags:
        include_tags will check for the tag
    :param ignore_tags:
        include_tags will check for the tag
    :return bool:
        The output will be boolean, based on the tag filter given
    """
    # Rule tag is not in the include tags list, return false
    if include_tags and rule_tags and set(include_tags).isdisjoint(set(rule_tags.get(finding.rule_name, []))):
        return False

    # Rule tag is in the ignore tags list, return false
    if ignore_tags and rule_tags and not set(ignore_tags).isdisjoint(set(rule_tags.get(finding.rule_name, []))):
        return False

    return True



# Standard Library
import argparse
import os


class EnvDefault(argparse.Action):
    """
    Helper for the CLI argparse to allow setting defaults through environment variables
    Usage: In an argparse argument, set the Action to this class.
        Add the extra variable envvar added that has the name of the environment variable containing the default value.
    Example: parser.add_argument("--gitleaks-path", action=EnvDefault, envvar="RESC_GITLEAKS_PATH")
        This would result in the parser reading the env var if it exists and using it as the default,
        always to be overrideable using the cli argument.
    """

    def __init__(self, envvar, required=True, default=None, **kwargs):
        if not default and envvar:
            if envvar in os.environ:
                default = os.environ[envvar]
        if required and default:
            required = False
        super().__init__(default=default, required=required, **kwargs)

    def __call__(self, parser, namespace, values, option_string=None):
        setattr(namespace, self.dest, values)



# pylint: disable=R0903
# First Party
from vcs_scanner.helpers.environment_wrapper import EnvironmentVariable

DEBUG_MODE = "DEBUG_MODE"

GITLEAKS_PATH = "GITLEAKS_PATH"

RABBITMQ_SERVICE_HOST = "RESC_RABBITMQ_SERVICE_HOST"
RABBITMQ_DEFAULT_VHOST = "RABBITMQ_DEFAULT_VHOST"
RABBITMQ_USERNAME = "RABBITMQ_USERNAME"
RABBITMQ_PASSWORD = "RABBITMQ_PASSWORD"
RABBITMQ_QUEUE = "RABBITMQ_QUEUE"

VCS_INSTANCES_FILE_PATH = "VCS_INSTANCES_FILE_PATH"

RESC_API_NO_AUTH_SERVICE_HOST = "RESC_API_NO_AUTH_SERVICE_HOST"
RESC_API_NO_AUTH_SERVICE_PORT = "RESC_API_NO_AUTH_SERVICE_PORT"

RESC_INCLUDE_TAGS = "RESC_INCLUDE_TAGS"
RESC_IGNORE_TAGS = "RESC_IGNORE_TAGS"

REQUIRED_ENV_VARS = [
    EnvironmentVariable(
        GITLEAKS_PATH,
        "The filepath for the gitleaks binary",
        required=True,
    ),
    EnvironmentVariable(
        RABBITMQ_SERVICE_HOST,
        "The hostname/IP address of the rabbitmq server.",
        required=True,
    ),
    EnvironmentVariable(
        RABBITMQ_DEFAULT_VHOST,
        "The virtual host name of the rabbitmq server.",
        required=True,
    ),
    EnvironmentVariable(
        DEBUG_MODE,
        "Show debug log statements, if set to '0' only INFO logs and above will be shown.",
        required=False,
        default="0",
    ),
    EnvironmentVariable(
        RESC_API_NO_AUTH_SERVICE_HOST,
        "URL of the RESC API authentication disabled endpoint",
        required=True,
    ),
    EnvironmentVariable(
        RESC_API_NO_AUTH_SERVICE_PORT,
        "PORT of the RESC API authentication disabled endpoint",
        default="8000",
    ),
    EnvironmentVariable(
        RABBITMQ_USERNAME,
        "The username used to connect to the rabbitmq queues.",
        required=True,
    ),
    EnvironmentVariable(
        RABBITMQ_PASSWORD,
        "The password used to connect to the rabbitmq queues.",
        required=True,
    ),
    EnvironmentVariable(
        RABBITMQ_QUEUE,
        "The rabbitmq queue to connect to.",
        required=True,
    ),
    EnvironmentVariable(
        VCS_INSTANCES_FILE_PATH,
        "The absolute path to the json file containing the vcs_instances_definitions",
        required=True,
    ),
    EnvironmentVariable(
        RESC_INCLUDE_TAGS,
        "Filter for outputting findings based on specified tags. Provided as comma seperated list.",
        required=False,
        default=None,
    ),
    EnvironmentVariable(
        RESC_IGNORE_TAGS,
        "Filter for NOT outputting findings based on specified tags. Provided as comma seperated list.",
        required=False,
        default=None,
    ),
]



# pylint: disable=no-name-in-module
# Standard Library
import os

# Third Party
from pydantic import BaseModel, conint, constr, validator
from resc_backend.resc_web_service.schema.repository import Repository
from resc_backend.resc_web_service.schema.vcs_provider import VCSProviders


class RepositoryRuntime(BaseModel):
    repository_name: str
    repository_id: str
    repository_url: str
    project_key: str
    vcs_instance_name: str
    latest_commit: str | None = None

    def convert_to_repository(self, vcs_instance_id: int) -> Repository:
        return Repository(
            project_key=self.project_key,
            repository_id=self.repository_id,
            repository_name=self.repository_name,
            repository_url=self.repository_url,
            vcs_instance=vcs_instance_id,
            latest_commit=self.latest_commit,
        )


class VCSInstanceRuntime(BaseModel):
    id_: int | None
    name: constr(max_length=200)
    provider_type: VCSProviders
    hostname: constr(max_length=200)
    port: conint(gt=-0, lt=65536)
    scheme: str
    username: constr(max_length=200)
    token: constr(max_length=200)
    exceptions: list[str] | None = []
    scope: list[str] | None = []
    organization: str | None

    @validator("scheme", pre=True)
    @classmethod
    def check_scheme(cls, value):
        allowed_schemes = ["http", "https"]
        if value not in allowed_schemes:
            raise ValueError(f"The scheme '{value}' must be one of the following {', '.join(allowed_schemes)}")
        return value

    @validator("organization", pre=True)
    @classmethod
    def check_organization(cls, value, values):
        if not value:
            if values["provider_type"] == VCSProviders.AZURE_DEVOPS:
                raise ValueError("The organization field needs to be specified for Azure devops vcs instances")
        return value

    @validator("scope", pre=True)
    @classmethod
    def check_scope_and_exceptions(cls, value, values):
        if value and values["exceptions"]:
            raise ValueError(
                "You cannot specify bot the scope and exceptions to the scan, only one setting" " is supported."
            )
        return value

    @validator("username", pre=True)
    @classmethod
    def check_presence_of_username(cls, value, values):
        if not os.environ.get(value):
            raise ValueError(
                f"The username for VCS Instance {values['name']} could not be found in the "
                f"environment variable {value}"
            )
        return os.environ.get(value)

    @validator("token", pre=True)
    @classmethod
    def check_presence_of_token(cls, value, values):
        if not os.environ.get(value):
            raise ValueError(
                f"The access token for VCS Instance {values['name']} could not be found in the "
                f"environment variable {value}"
            )
        return os.environ.get(value)



# pylint: disable=E1101
# Standard Library
import csv
import logging
from datetime import UTC, datetime

# Third Party

logger = logging.getLogger(__name__)


class IgnoredListProvider:  # pylint: disable=R0902
    def __init__(self, ignore_findings_path: str | None):
        self.ignore_findings_path: str | None = ignore_findings_path
        self.today: datetime = datetime.now(UTC)

    def get_ignore_list(self) -> dict[str, True]:
        """
        Get the dictionary of ignored findings according to the file
        The output will contain a dictionary with the path|rule|line as key and True as value.
        We use a dictionary for random access instead of list.
        """
        if self.ignore_findings_path is None:
            return {}

        ignored = {}

        try:
            # read dsv: `path|rule_name|line_number|expiry_date`
            with open(self.ignore_findings_path, encoding="utf-8") as ignore_findings_file:
                csv_ignore_list = csv.reader(ignore_findings_file, delimiter="|")
                for row in csv_ignore_list:
                    if self._is_row_comment(row):
                        continue

                    if not self._is_row_valid(row):
                        continue

                    if not self._is_row_active(row):
                        continue

                    # we use the path, rule_name, line_number as a dictionary key
                    path = row[0]
                    rule = row[1]
                    line = row[2]
                    ignored[f"{path}|{rule}|{line}"] = True
        except FileNotFoundError:  # <- File does not exists: we just fail silently
            logger.warning(f"could not find {self.ignore_findings_path}")
            return {}

        return ignored

    def _is_row_comment(self, row: list[str]) -> bool:
        """
        Line starting with # are comments
        """
        return row[0][:1] == "#"

    def _is_row_valid(self, row: list[str]) -> bool:
        """
        If a line is strictly shorter than 3, it is not a valid entry
        """
        if len(row) >= 3:
            return True

        string_row: str = "".join(row)
        logger.warning(f"Skipping: incomplete entry for {string_row}")
        return False

    def _is_row_active(self, row: list[str]) -> bool:
        """
        A row is active if it is of length 3 or if the date of the row is not in the past.
        """
        if len(row) == 3:
            return True

        path = row[0]
        date = row[3]
        try:
            expire: datetime = datetime.fromisoformat(date).replace(tzinfo=UTC)

        except ValueError:
            logger.warning(f"Skipping: invalid date entry for {path}: {date}")
            return False

        if expire < self.today:
            logger.warning(f"Info: expired date entry for {date}")
            return False

        return True



# Standard Library
import logging.config
import sys
import sysconfig
from os import path

# Third Party
import tomlkit

# First Party
from vcs_scanner.input_parser import parse_vcs_instances_file
from vcs_scanner.model import VCSInstanceRuntime

logger = logging.getLogger(__name__)


def get_logging_settings_path():
    if path.isfile(sysconfig.get_path("purelib") + "/vcs_scanner"):
        base_dir = sysconfig.get_path("purelib") + "/vcs_scanner"
    else:
        base_dir = path.dirname(__file__)

    return base_dir + "/static/logging.ini"


def generate_logger_config(log_file_path, debug=True):
    """A function to generate the global logger config dictionary

    Arguments:
        log_file_path {string} -- Path where the logs are to be stored

    Keyword Arguments:
        debug {bool} -- Whether the logging level should be set to DEBUG or INFO (default: {True})

    Returns:
        Dict -- A dictionary containing the logger configuration
    """

    logging_level = "DEBUG" if debug else "INFO"
    logging_config = {
        "version": 1,
        "disable_existing_loggers": False,
        "formatters": {
            "generic-log-formatter": {"format": "[%(levelname)s] [%(name)s] [%(asctime)s] %(message)s"},
        },
        "handlers": {
            "console": {
                "level": logging_level,
                "class": "logging.StreamHandler",
                "formatter": "generic-log-formatter",
            },
            "file": {
                "level": logging_level,
                "class": "logging.handlers.RotatingFileHandler",
                "formatter": "generic-log-formatter",
                "filename": log_file_path,
                "maxBytes": 100 * 1024 * 1024,
                "backupCount": 5,
            },
        },
        "loggers": {
            "": {
                "handlers": ["console", "file"],
                "level": logging_level,
                "propagate": True,
            },
        },
    }

    return logging_config


def initialise_logs(log_file_path: str, debug=True):
    logging_ini_file = get_logging_settings_path()
    logging.config.fileConfig(
        logging_ini_file,
        defaults={"log_file_path": log_file_path},
        disable_existing_loggers=False,
    )
    logger_config = logging.getLogger("root")
    if int(debug) == 1:
        logger_config.setLevel(logging.DEBUG)
    else:
        logger_config.setLevel(logging.INFO)
    return logger_config


def load_vcs_instances(file_path: str) -> dict[str, VCSInstanceRuntime]:
    vcs_instances_list: list[VCSInstanceRuntime] = parse_vcs_instances_file(file_path)
    if not vcs_instances_list:
        logger.critical(f"Exiting due to issues in VCS Instances definition in file {file_path}")
        sys.exit(-1)
    vcs_instances_map: dict[str, VCSInstanceRuntime] = {
        vcs_instance.name: vcs_instance for vcs_instance in vcs_instances_list
    }
    return vcs_instances_map


def get_rule_pack_version_from_file(file_content: str) -> str | None:
    toml_rule_dictionary = tomlkit.loads(file_content)
    rule_pack_version = toml_rule_dictionary.get("version", None)
    return rule_pack_version



# pylint: disable=W0212
# Standard Library
import json
import logging
import sys
from argparse import Namespace
from datetime import datetime
from pathlib import Path

# Third Party
from resc_backend.constants import TEMP_RULE_FILE
from resc_backend.resc_web_service.schema.finding import FindingBase, FindingCreate
from resc_backend.resc_web_service.schema.repository import (
    RepositoryCreate,
    RepositoryRead,
)
from resc_backend.resc_web_service.schema.scan import Scan, ScanCreate, ScanRead
from resc_backend.resc_web_service.schema.scan_type import ScanType
from resc_backend.resc_web_service.schema.vcs_instance import (
    VCSInstanceCreate,
    VCSInstanceRead,
)
from resc_backend.resc_web_service_interface.findings import (
    create_findings_with_scan_id,
)
from resc_backend.resc_web_service_interface.repositories import (
    create_repository,
    get_last_scan_for_repository,
)
from resc_backend.resc_web_service_interface.rule_packs import (
    download_rule_pack_toml_file,
    get_rule_packs,
)
from resc_backend.resc_web_service_interface.scans import create_scan
from resc_backend.resc_web_service_interface.vcs_instances import create_vcs_instance
from tenacity import retry, stop_after_attempt, wait_exponential

# First Party
from vcs_scanner.common import get_rule_pack_version_from_file
from vcs_scanner.helpers.finding_filter import should_process_finding
from vcs_scanner.helpers.providers.rule_tag import RuleTagProvider
from vcs_scanner.model import VCSInstanceRuntime
from vcs_scanner.output_modules.output_module import OutputModule

logger = logging.getLogger(__name__)


class RESTAPIWriter(OutputModule):
    def __init__(
        self,
        rws_url,
        ignore_tags: list[str] = None,
        include_tags: list[str] = None,
        rule_tag_provider: RuleTagProvider = RuleTagProvider(),
    ):
        self.rws_url = rws_url
        self.ignore_tags = ignore_tags
        self.include_tags = include_tags
        self.rule_tag_provider: RuleTagProvider = rule_tag_provider

    def load_rules(self, toml_rule_file_path: str) -> None:
        self.rule_tag_provider.load(toml_rule_file_path)

    def write_vcs_instance(self, vcs_instance_runtime: VCSInstanceRuntime) -> VCSInstanceRead | None:
        created_vcs_instance = None
        vcs_instance = VCSInstanceCreate(
            name=vcs_instance_runtime.name,
            provider_type=vcs_instance_runtime.provider_type,
            hostname=vcs_instance_runtime.hostname,
            port=vcs_instance_runtime.port,
            scheme=vcs_instance_runtime.scheme,
            exceptions=vcs_instance_runtime.exceptions,
            scope=vcs_instance_runtime.scope,
            organization=vcs_instance_runtime.organization,
        )
        response = create_vcs_instance(self.rws_url, vcs_instance)
        if response.status_code == 201:
            created_vcs_instance = VCSInstanceRead(**json.loads(response.text))
        else:
            logger.warning(f"Creating vcs_instance failed with error: {response.status_code}->{response.text}")
        return created_vcs_instance

    def write_repository(self, repository: RepositoryCreate) -> RepositoryRead | None:
        created_repository = None
        response = create_repository(self.rws_url, repository)
        if response.status_code == 201:
            created_repository = RepositoryRead(**json.loads(response.text))
        else:
            logger.warning(f"Creating repository failed with error: {response.status_code}->{response.text}")
        return created_repository

    def write_findings(
        self,
        scan_id: int,
        repository_id: int,
        scan_findings: list[FindingBase],
        repository_name: str = "",
    ) -> None:
        findings_create = []

        rule_tags = self.rule_tag_provider.get_rule_tags()
        for finding in scan_findings:
            # We strip the repository name here because in the case of
            # scan as dir the path of the finding is prefixed with the repository name
            if finding.author == "vcs-scanner":
                finding.file_path = finding.file_path.removeprefix(repository_name + "/")

            new_finding = FindingCreate.create_from_base_class(base_object=finding, repository_id=repository_id)

            if should_process_finding(
                finding=finding,
                rule_tags=rule_tags,
                ignore_tags=self.ignore_tags,
                include_tags=self.include_tags,
            ):
                findings_create.append(new_finding)

        response = create_findings_with_scan_id(self.rws_url, findings_create, scan_id)

        if response.status_code != 201:
            logger.warning(
                f"Creating findings for scan {scan_id} " f"failed with error: {response.status_code}->{response.text}"
            )
        logger.info(f"Found {len(scan_findings)} issues during scan for scan_id: {scan_id} ")

    def write_scan(
        self,
        scan_type_to_run: ScanType,
        last_scanned_commit: str,
        scan_timestamp: datetime,
        repository: RepositoryRead,
        rule_pack: str,
    ) -> ScanRead | None:
        created_scan = None
        scan_object = ScanCreate.create_from_base_class(
            base_object=Scan(
                scan_type=scan_type_to_run,
                last_scanned_commit=last_scanned_commit,
                timestamp=scan_timestamp,
                rule_pack=rule_pack,
            ),
            repository_id=repository.id_,
        )

        response = create_scan(self.rws_url, scan_object)
        if response.status_code == 201:
            created_scan = ScanRead(**json.loads(response.text))
            logger.info(f"Successfully created scan for repository {repository.repository_url} ")
        else:
            logger.warning(
                f"Creating {scan_type_to_run} scan failed with error: {response.status_code}->{response.text}"
            )

        return created_scan

    def get_last_scan_for_repository(self, repository: RepositoryRead) -> ScanRead | None:
        response = get_last_scan_for_repository(self.rws_url, repository.id_)
        if not response.status_code == 200:
            logger.warning(f"Retrieving last scan details failed with error: {response.status_code}->{response.text}")
            return None

        if not response.text or response.text == "null":
            return None
        return ScanRead(**json.loads(response.text))

    @retry(wait=wait_exponential(multiplier=1, min=2, max=10), stop=stop_after_attempt(100))
    def write_vcs_instances(self, vcs_instances_dict: dict[str, VCSInstanceRuntime]) -> dict[str, VCSInstanceRuntime]:
        try:
            for key in vcs_instances_dict:
                vcs_instance = vcs_instances_dict[key]
                vcs_instance_created = self.write_vcs_instance(vcs_instance)
                if not vcs_instance_created:
                    raise ValueError(f"Failed creating vcs instance {vcs_instance.name}")
                vcs_instance.id_ = vcs_instance_created.id_
                vcs_instances_dict[key] = vcs_instance
            return vcs_instances_dict
        except ValueError as ex:
            logger.error(f"Failed creating vcs instances, is the API available? | {ex} | Retrying...")
            raise

    def get_active_rule_pack_version(self) -> str | None:
        """
            Retrieve active rule pack version from database
        :return: str
            Return active rule pack version
        """
        active_rule_pack_version = None
        response = get_rule_packs(url=self.rws_url, active=True)
        if response.status_code == 200:
            json_body = json.loads(response.text)
            active_rule_pack_version = json_body["data"][0]["version"] if json_body else None
        else:
            logger.warning(
                f"Retrieving active rule pack version failed with error: {response.status_code}->{response.text}"
            )
        return active_rule_pack_version

    def download_rule_pack(self, rule_pack_version: str | None = "") -> str | None:
        """
            Download rule pack
        :param rule_pack_version:
            optional, filter on rule pack version
        :return: str
            Return downloaded rule pack version
        """
        response = download_rule_pack_toml_file(self.rws_url, rule_pack_version)
        if not response.status_code == 200:
            logger.error(
                f"Aborting scan! Downloading rule pack version {rule_pack_version} failed with "
                f"error: {response.status_code}->{response.text}"
            )
            sys.exit(-1)

        filename = Path(TEMP_RULE_FILE)
        filename.write_bytes(response.content)
        if rule_pack_version:
            logger.debug(
                f"Rule pack version: {rule_pack_version} has been successfully "
                f"downloaded to location {TEMP_RULE_FILE}"
            )
        else:
            rule_pack_version = get_rule_pack_version_from_file(response.content)
            if not rule_pack_version:
                logger.warning("Unable to obtain the rule pack version from downloaded file, defaulting to '0.0.0'")
            logger.debug(
                f"Latest rule pack version: {rule_pack_version} has been successfully "
                f"downloaded to location {TEMP_RULE_FILE}"
            )
        return rule_pack_version

    def check_active_rule_pack_version(self, rule_pack_version: str = None) -> str:
        """
            Check active rule pack version
        :return: str
            Return active rule pack version
        """
        if rule_pack_version:
            rule_pack_version_from_db = self.get_active_rule_pack_version()

            if rule_pack_version != rule_pack_version_from_db:
                rule_pack_version = self.download_rule_pack(rule_pack_version_from_db)
        else:
            rule_pack_version = self.download_rule_pack()
        return rule_pack_version

    @staticmethod
    def make(args: Namespace) -> "RESTAPIWriter":
        """
            Get the Rest API writer given the args provided.

        :param args:
            Namespace object containing the CLI arguments
        """
        rule_tag_provider = RuleTagProvider()
        rule_tag_provider.load(args.gitleaks_rules_path)

        output_plugin = RESTAPIWriter(
            rws_url=args.rws_url,
            include_tags=args.include_tags,
            ignore_tags=args.ignore_tags,
            rule_tag_provider=rule_tag_provider,
        )
        return output_plugin



# Standard Library
import json
import logging
import subprocess
from datetime import UTC, datetime

# Third Party
from resc_backend.resc_web_service.schema.finding import FindingBase

# First Party
from vcs_scanner.constants import LEAKS_FOUND_EXIT_CODE, NO_LEAKS_FOUND_EXIT_CODE

logger: logging.Logger = logging.getLogger(__name__)


class GitLeaksWrapper:
    SCAN_TMP_DIRECTORY: str = "."

    def __init__(
        self,
        rules_filepath: str,
        report_filepath: str,
        repository_path: str,
        scan_from: str = None,
        gitleaks_path: str = "gitleaks",
        git_scan: bool = True,
    ):
        self.rules_filepath = rules_filepath
        self.report_filepath = report_filepath
        self.repository_path = repository_path
        self.scan_from = scan_from
        self.gitleaks_path = gitleaks_path
        self.git_scan = git_scan

    def _build_gitleaks_command(self):
        # Base scan command
        command = [
            f"{self.gitleaks_path}",
            "detect",
            f"--source={self.repository_path}",
            f"--config={self.rules_filepath}",
            f"--report-path={self.report_filepath}",
            f"--exit-code={LEAKS_FOUND_EXIT_CODE}",
        ]

        if not self.git_scan:
            command.append("--no-git")

        # Incremental scan command
        if self.scan_from:
            command.append(f"--log-opts={self.scan_from}..")
        return command

    def start_scan(self) -> list[FindingBase]:
        """
        :return: Output.
            If Successful, a list of FindingCreate objects is returned.
            Otherwise, a None object is returned
        """
        try:
            result = subprocess.run(
                self._build_gitleaks_command(),
                check=False,
                capture_output=True,
            )

            exitcode = result.returncode
            if exitcode == NO_LEAKS_FOUND_EXIT_CODE:
                return []
            if exitcode == LEAKS_FOUND_EXIT_CODE:
                return self._parse_output(self.report_filepath)

            error_output = result.stderr.decode("utf-8")
            logger.error(f"GitLeaks exited with an unexpected code: {exitcode}. Output: {error_output}")
            return []

        except subprocess.CalledProcessError as called_process_error:
            logger.error(
                f"Error encountered while running the gitleaks process: {called_process_error.stdout.decode('utf-8')}"
            )
            return []
        except FileNotFoundError as error:
            logger.error(f"Unable to locate a file: {error}")
            return []

    @staticmethod
    def _calculate_permanent_leak_url(leak_url: str, repository: str, commit_id: str) -> str:
        """
            The Leak URL given by bitbucket is not correct.
            Construct a proper one that links to the exact change in the commit
        :param leak_url: The url of the leak (Bitbucket format)
        :param repository: The git repository
        :param commit_id: The commit hash
        :return: the permalink to the leak
        """
        new_url = leak_url

        # Remove the line-number from the url.
        hash_pos = new_url.rfind("#")
        if hash_pos > 0:
            new_url = new_url[:hash_pos]

        new_url = new_url.replace("/scm/", "/projects/")
        new_url = new_url.replace("/blob/", "/commits/")
        new_url = new_url.replace(f"/{repository}/", f"/repos/{repository[:-4]}/")  # remove .git at the end
        new_url = new_url.replace(f"/{commit_id}/", f"/{commit_id}#")

        return new_url

    @staticmethod
    def _get_valid_timestamp(timestamp: str) -> datetime:
        if timestamp == "":
            logger.debug("Date is empty, using now(UTC) instead.")
            return datetime.now(UTC)

        try:
            converted_timestamp: datetime | None = datetime.fromisoformat(timestamp)
        except ValueError:
            logger.debug(f"{timestamp} has an unexpected date format. Expected ISO 8601")
            converted_timestamp = datetime.now(UTC)

        return converted_timestamp

    @classmethod
    def _parse_output(cls, file_path: str) -> list[FindingBase]:
        """
        Parse the gitleaks findings from the temp file and return a list of Finding objects
        :param file_path: the tempfile containing the gitleaks findings
        :return: list of Finding objects
        """
        findings = []
        with open(file_path, encoding="utf-8") as report_file:
            results = json.load(report_file)

        for result in results:
            commit_timestamp = cls._get_valid_timestamp(result["Date"])
            finding = FindingBase(
                file_path=result["File"],
                line_number=result["StartLine"],
                column_start=result["StartColumn"],
                column_end=result["EndColumn"],
                email=result["Email"],
                author=result["Author"],
                commit_id=result["Commit"],
                commit_message=result["Message"],
                commit_timestamp=commit_timestamp,
                rule_name=result["RuleID"],
            )
            logger.debug(finding)

            findings.append(finding)

        return findings



# pylint: disable=E1101
# Standard Library
import logging
import os
import shutil
import time
import uuid
from collections.abc import Callable
from datetime import UTC, datetime

# Third Party
from git import Commit
from resc_backend.resc_web_service.schema.finding import FindingBase
from resc_backend.resc_web_service.schema.repository import Repository, RepositoryBase
from resc_backend.resc_web_service.schema.scan import Scan, ScanRead
from resc_backend.resc_web_service.schema.scan_type import ScanType

# First Party
from vcs_scanner.helpers.providers.rule_file import RuleFileProvider
from vcs_scanner.output_modules.output_module import OutputModule
from vcs_scanner.resc_worker import RESCWorker
from vcs_scanner.secret_scanners.git_operation import clone_repository
from vcs_scanner.secret_scanners.gitleaks_wrapper import GitLeaksWrapper

# This is an arbitrary number to distinguish between no issues, an error and
# the situation in which leaks are found. Note that this number cannot be bigger than 255 (OS limitation)
LEAKS_FOUND_EXIT_CODE = 42
NO_LEAKS_FOUND_EXIT_CODE = 0

logger = logging.getLogger(__name__)


class SecretScanner(RESCWorker):  # pylint: disable=R0902
    def __init__(
        self,
        gitleaks_binary_path: str,
        gitleaks_rules_provider: RuleFileProvider,
        rule_pack_version: str,
        output_plugin: OutputModule,
        repository: Repository,
        username: str,
        personal_access_token: str,
        scan_tmp_directory: str = ".",
        local_path: str | None = None,
        force_base_scan: bool = False,
        latest_commit: str | None = None,
    ):
        self.rule_provider: RuleFileProvider | None = None
        self.gitleaks_rules_provider: RuleFileProvider = gitleaks_rules_provider
        self.gitleaks_binary_path: str = gitleaks_binary_path
        self.rule_pack_version: str = rule_pack_version
        self._output_module: OutputModule = output_plugin
        self._scan_tmp_directory: str = scan_tmp_directory
        self.repository: Repository = repository
        self.username: str = username
        self.personal_access_token: str = personal_access_token
        self.local_path = local_path
        self.force_base_scan = force_base_scan
        self.latest_commit = latest_commit
        self.head_commit: None | Commit = None

        self._as_dir: bool = False
        self._as_repo: bool = False
        self._created_repository: None | RepositoryBase = None
        self._last_scanned_commit: None | str = None
        self._scan_type_to_run: None | ScanType = None
        self._scan_timestamp_start: None | datetime = None
        self._created_scan: None | ScanRead = None
        self._repo_clone_path: None | str = None
        self._findings_from_repo: list[FindingBase] = []
        self._findings_from_dir: list[FindingBase] = []
        self._findings: list[FindingBase] = []

        if self.local_path:
            self.repo_display_name = self.local_path.replace(".", "_").replace("/", "_")
        else:
            self.repo_display_name = self.repository.repository_url

    def run_scan(self, as_dir: bool = False, as_repo: bool = False) -> None:
        """
        Run the scan steps by steps.

        Args:
            as_dir (bool, optional): whether we scan as directory. Defaults to False.
            as_repo (bool, optional): whether we scan as repository. Defaults to False.
        """
        self._as_dir = as_dir
        self._as_repo = as_repo

        pipes: list[Callable[[], bool]] = [
            self._is_valid,
            self._is_scan_needed_from_latest_commit,
            self._create_repository,
            self._fetch_last_scanned_commit,
            self._is_scan_needed,
            self._start_timer,
            self._create_scan,
            self._clone_repo,
            self._run_repo_scan,
            self._run_dir_scan,
            self._merge_findings,
            self._write_findings,
            self._cleaning_up,
        ]

        for pipe in pipes:
            # If the pipe does not succeed we exit immediately.
            if not pipe():
                return

    def _is_valid(self) -> bool:
        if not self._as_dir and not self._as_repo:
            logger.error("no scan type selected")
            return False
        return True

    def _is_scan_needed_from_latest_commit(self) -> bool:
        if self._as_repo and not self.latest_commit:
            # There is no latest commit for this repository, assuming that its empty
            logger.info(
                f"Skipping scanning of {self.repository.project_key}/{self.repository.repository_name} "
                f"there are no commits"
            )
            return False
        logger.info(
            f"Started task for scanning {self.repository.repository_name} using "
            f"rule pack version: {self.rule_pack_version}"
        )
        return True

    def _create_repository(self) -> bool:
        # Insert in to repository table
        self._created_repository = self._output_module.write_repository(self.repository)
        if not self._created_repository:
            logger.error(
                f"Error processing "
                f"{self.repository.repository_name}."
                f" Error details: unable to create repository: {self._created_repository}"
            )
            return False

        logger.info(f"Scanning repository {self.repository.project_key}/{self.repository.repository_name}")
        return True

    def _fetch_last_scanned_commit(self) -> True:
        # Get last scanned commit for the repository
        last_scan_for_repository = self._output_module.get_last_scan_for_repository(repository=self._created_repository)
        self._last_scanned_commit = last_scan_for_repository.last_scanned_commit if last_scan_for_repository else None
        self._scan_type_to_run = self._determine_scan_type(
            last_scan_for_repository=last_scan_for_repository,
        )
        return True

    def _determine_scan_type(self, last_scan_for_repository: Scan) -> ScanType | None:
        # Force base scan, or has no previous scan
        if self.force_base_scan or last_scan_for_repository is None:
            return ScanType.BASE
        # Has previous scan
        if last_scan_for_repository:
            # Rule-pack is different from previous scan
            if last_scan_for_repository.rule_pack != self.rule_pack_version:
                return ScanType.BASE
            # Last commit is different from previous scan
            if self.latest_commit and self.latest_commit != last_scan_for_repository.last_scanned_commit:
                return ScanType.INCREMENTAL
        # Skip scanning, no conditions match
        return None

    def _is_scan_needed(self) -> bool:
        if self._scan_type_to_run is None:
            logger.info(
                "Skipped scanning on repository: "
                f"{self.repository.project_key}/{self.repository.repository_name} no new commits found."
            )
            return False
        return True

    def _start_timer(self) -> True:
        self._scan_timestamp_start = datetime.now(UTC)
        return True

    def _create_scan(self) -> bool:
        self._created_scan = self._output_module.write_scan(
            self._scan_type_to_run,
            self.latest_commit,
            self._scan_timestamp_start.isoformat(),
            self._created_repository,
            rule_pack=self.rule_pack_version,
        )
        if not self._created_scan:
            logger.error(
                f"Error processing {self.repository.project_key}/{self.repository.repository_name} "
                f"Error details: unable to create scan object"
            )
            return False
        return True

    def _clone_repo(self) -> True:
        # Clone and run scan upon the repository
        if not self.local_path:
            self._repo_clone_path = f"{self._scan_tmp_directory}/{self.repository.repository_name}"
            self.head_commit = clone_repository(
                repository_url=self.repository.repository_url,
                repo_clone_path=self._repo_clone_path,
                username=self.username,
                personal_access_token=self.personal_access_token,
            )
        else:
            self._repo_clone_path = self.local_path
        return True

    def _run_repo_scan(self) -> True:
        if not self._as_repo:
            return True

        if self.gitleaks_rules_provider.scan_as_repo_rule_file_path is None:
            return True

        logger.info(
            f"Started task for scanning {self._repo_clone_path} using rule pack version: {self.rule_pack_version}"
        )
        scan_timestamp_start = datetime.now(UTC)
        self._findings_from_repo = self._scan_repo(self._scan_type_to_run, self._last_scanned_commit)
        scan_timestamp_end = datetime.now(UTC)
        logger.info(
            f"Running {self._scan_type_to_run} scan on repository "
            f"{self.repository.project_key}/{self.repository.repository_name}"
            f" took {scan_timestamp_end - scan_timestamp_start} ms."
        )
        return True

    def _scan_repo(self, scan_type_to_run: str, last_scanned_commit: str) -> list[FindingBase]:
        """
            Clone and scan the given repository
        :param scan_type_to_run:
            Type of scan to run (Base or Incremental)
        :param last_scanned_commit:
            Last scanned commit of the repository to scan
        :return: Success, output.
            If Success is False, the output will contain an error message.
            Otherwise, the output will contain a list of findings or an empty list if no issue was found
        """

        logger.debug(f"Started scanning {self.repo_display_name}")
        if not self.local_path:
            report_filepath = f"{self._scan_tmp_directory}/{self._repo_clone_path}_{str(uuid.uuid4().hex)}.json"
        else:
            report_filepath = f"{self.local_path}/{self.repo_display_name}_{str(uuid.uuid4().hex)}.json"
        try:
            if scan_type_to_run == ScanType.BASE:
                scan_from = None
            elif scan_type_to_run == ScanType.INCREMENTAL and last_scanned_commit:
                scan_from = last_scanned_commit
            else:
                scan_from = None

            gitleaks_command = GitLeaksWrapper(
                scan_from=scan_from,
                gitleaks_path=self.gitleaks_binary_path,
                repository_path=self._repo_clone_path,
                rules_filepath=self.gitleaks_rules_provider.scan_as_repo_rule_file_path,
                report_filepath=report_filepath,
            )

            before_scan = time.time()
            findings: list[FindingBase] = gitleaks_command.start_scan()
            after_scan = time.time()
            scan_duration = int(after_scan - before_scan)
            logger.info(f"scan of repository {self._repo_clone_path} took {scan_duration} seconds")
            return findings
        except BaseException as error:
            logger.error(
                f"An exception occurred while scanning repository {self.repository.repository_url} " f"error: {error}"
            )
            return []
        finally:
            # Make sure the tempfile and repo cloned path removed
            logger.debug(f"Cleaning up the temporary report: {report_filepath}")
            if os.path.exists(report_filepath):
                os.remove(report_filepath)

    def _run_dir_scan(self):
        if not self._as_dir:
            return True

        if self.gitleaks_rules_provider.scan_as_dir_rule_file_path is None:
            return True

        logger.info(
            f"Started task for scanning {self._repo_clone_path} using rule pack version: {self.rule_pack_version}"
        )

        scan_timestamp_start = datetime.now(UTC)
        self._findings_from_dir = self._scan_directory(self._repo_clone_path)
        scan_timestamp_end = datetime.now(UTC)
        logger.info(
            f"Running directory scan on {self._repo_clone_path}"
            f" took {scan_timestamp_end - scan_timestamp_start} ms."
        )
        return True

    def _scan_directory(self, directory_path: str) -> list[FindingBase] | None:
        """
            Scan the given directory
        :param directory_path:
            Directory path to be scanned
        :return: Optional[List[FindingBase]].
            The output will contain a list of findings or an empty list if no finding was found
        """
        logger.debug(f"Started scanning {self.repo_display_name}:{directory_path}")
        if not self.local_path:
            report_filepath = f"{self._scan_tmp_directory}/{directory_path}_{str(uuid.uuid4().hex)}.json"
        else:
            report_filepath = f"{self.local_path}/{self.repo_display_name}_{str(uuid.uuid4().hex)}.json"
        try:
            gitleaks_command = GitLeaksWrapper(
                scan_from=None,
                gitleaks_path=self.gitleaks_binary_path,
                repository_path=directory_path,
                rules_filepath=self.gitleaks_rules_provider.scan_as_dir_rule_file_path,
                report_filepath=report_filepath,
                git_scan=False,
            )

            before_scan = time.time()
            findings: list[FindingBase] = gitleaks_command.start_scan()
            after_scan = time.time()
            scan_duration = int(after_scan - before_scan)
            logger.info(f"scan of repository {directory_path} took {scan_duration} seconds")
            return findings
        except BaseException as error:
            logger.error(f"An exception occurred while scanning directory {directory_path} error: {error}")
        finally:
            # Make sure the tempfile is removed
            logger.debug(f"Cleaning up the temporary report: {report_filepath}")
            if os.path.exists(report_filepath):
                os.remove(report_filepath)
        return None

    def _merge_findings(self) -> bool:
        if len(self._findings_from_dir) == 0 and len(self._findings_from_repo) == 0:
            path = (
                self.local_path
                if self.local_path
                else self.repository.project_key + "/" + self.repository.repository_name
            )
            logger.info(f"No findings registered in {path}.")
            return False

        self._findings = self._findings_from_repo + self._findings_from_dir
        self._findings = list(map(self._populate_if_empty, self._findings))
        return True

    def _populate_if_empty(self, finding: FindingBase) -> FindingBase:
        finding.commit_id = finding.commit_id or (
            self.head_commit.hexsha if self.head_commit is not None else "unknown"
        )
        finding.commit_message = finding.commit_message or (
            self.head_commit.message if self.head_commit is not None else ""
        )
        finding.commit_timestamp = finding.commit_timestamp or (
            self.head_commit.committed_date if self.head_commit is not None else datetime.now(UTC)
        )
        finding.author = finding.author or "vcs-scanner"
        return finding

    def _write_findings(self) -> True:
        logger.info(f"Scan completed: {len(self._findings)} findings were found.")
        self._output_module.write_findings(
            repository_id=getattr(self._created_repository, "id_", 0),
            scan_id=getattr(self._created_scan, "id_", 0),
            scan_findings=self._findings,
            repository_name=self.repository.repository_name,
        )
        return True

    def _cleaning_up(self) -> True:
        # Make sure the tempfile and repo cloned path removed
        if self._repo_clone_path and not self.local_path and os.path.exists(self._repo_clone_path):
            logger.debug(f"Cleaning up the repository cloned directory: {self._repo_clone_path}")
            shutil.rmtree(self._repo_clone_path)
        return True



import pathlib
from argparse import ArgumentParser

from vcs_scanner.helpers.env_default import EnvDefault


def create_cli_argparser() -> ArgumentParser:
    """
        Create ArgumentParser for CLI arguments
    :return: ArgumentParser.
        ArgumentParser instance with all arguments as expected for RESC
    """
    parser_common = ArgumentParser(add_help=False)
    parser_common.add_argument(
        "--gitleaks-path",
        type=pathlib.Path,
        action=EnvDefault,
        envvar="RESC_GITLEAKS_PATH",
        required=True,
        help="Path to the gitleaks binary. " "Can also be set via the RESC_GITLEAKS_PATH environment variable",
    )
    parser_common.add_argument(
        "--gitleaks-rules-path",
        type=pathlib.Path,
        action=EnvDefault,
        required=True,
        envvar="RESC_GITLEAKS_RULES_PATH",
        help="Path to the gitleaks rules file. "
        "Can also be set via the "
        "RESC_GITLEAKS_RULES_PATH environment variable",
    )
    parser_common.add_argument(
        "--ignored-blocker-path",
        type=pathlib.Path,
        action=EnvDefault,
        required=False,
        envvar="RESC_IGNORED_BLOCKER_PATH",
        help="Path to the resc-ignore.dsv file. "
        "Can also be set via the "
        "RESC_IGNORED_BLOCKER_PATH "
        "environment variable",
    )
    parser_common.add_argument(
        "-w",
        "--exit-code-warn",
        required=False,
        action=EnvDefault,
        default=2,
        type=int,
        envvar="RESC_EXIT_CODE_WARN",
        help="Exit code given if CLI encounters findings tagged with Warn, default 2. "
        "Can also be set via the RESC_EXIT_CODE_WARN environment variable",
    )
    parser_common.add_argument(
        "-b",
        "--exit-code-block",
        required=False,
        action=EnvDefault,
        default=1,
        type=int,
        envvar="RESC_EXIT_CODE_BLOCK",
        help="Exit code given if CLI encounters findings tagged with Block, default 1. "
        "Can also be set via the RESC_EXIT_CODE_BLOCK environment variable",
    )
    parser_common.add_argument(
        "--include-tags",
        required=False,
        action=EnvDefault,
        type=str,
        envvar="RESC_INCLUDE_TAGS",
        help="Filter for outputting findings based on specified tags. "
        "Provided as comma separated list. "
        "Can also be set via the RESC_INCLUDE_TAGS environment variable",
    )
    parser_common.add_argument(
        "--ignore-tags",
        required=False,
        action=EnvDefault,
        type=str,
        envvar="RESC_IGNORE_TAGS",
        help="Filter for NOT outputting findings based on specified tags. "
        "Provided as comma separated list. "
        "Can also be set via the RESC_IGNORE_TAGS environment variable",
    )
    parser_common.add_argument(
        "-v",
        "--verbose",
        required=False,
        action="store_true",
        help="Enable more verbose logging",
    )

    repository_common = ArgumentParser(add_help=False)
    repository_common.add_argument(
        "--repo-name",
        type=str,
        required=False,
        action=EnvDefault,
        envvar="RESC_REPO_NAME",
        help="The name of the repository. " "Can also be set via the RESC_REPO_NAME environment variable",
    )
    repository_common.add_argument("--force-base-scan", required=False, action="store_true")

    repository_common.add_argument(
        "--rws-url",
        type=str,
        required=False,
        action=EnvDefault,
        envvar="RESC_RWS_URL",
        help="The URL to the secret tracking service to which the scan results should "
        "be written. "
        "Can also be set via the RESC_RWS_URL environment variable",
    )

    parser: ArgumentParser = ArgumentParser()

    subparser = parser.add_subparsers(title="command", dest="command", required=True, help="Options dir, repo")
    directory = subparser.add_parser(
        "dir",
        description="Scan a directory",
        help="Scan a directory",
        parents=[parser_common],
    )
    repository = subparser.add_parser("repo", description="Scan a Git repository", help="Scan a Git repository")

    directory.add_argument(
        "--dir",
        type=pathlib.Path,
        required=True,
        action=EnvDefault,
        envvar="RESC_SCAN_PATH",
        help="The path to the directory where the scan target. "
        "Can also be set via the RESC_SCAN_PATH environment variable",
    )

    repository_subparser = repository.add_subparsers(
        title="repository_location",
        dest="repository_location",
        required=True,
        help="Options local, remote",
    )
    repository_local = repository_subparser.add_parser(
        "local",
        description="Scan a locally already cloned repository",
        help="Scan a locally already cloned repository",
        parents=[parser_common, repository_common],
    )
    repository_remote = repository_subparser.add_parser(
        "remote",
        description="Scan a remote repository",
        help="Scan a remote repository",
        parents=[parser_common, repository_common],
    )

    repository_local.add_argument(
        "--dir",
        type=pathlib.Path,
        required=True,
        action=EnvDefault,
        envvar="RESC_SCAN_PATH",
        help="The path to the directory where the repo is located. "
        "Can also be set via the RESC_SCAN_PATH environment variable",
    )

    repository_remote.add_argument(
        "--repo-url",
        type=str,
        required=True,
        action=EnvDefault,
        envvar="RESC_REPO_URL",
        help="url to repository you want to scan. " "Can also be set via the RESC_REPO_URL environment variable",
    )
    repository_remote.add_argument(
        "--username",
        type=str,
        required=False,
        action=EnvDefault,
        envvar="RESC_REPO_USERNAME",
        help="The username used for cloning the repository, "
        "you will be prompted for the password. "
        "Can also be set via the RESC_REPO_USERNAME & RESC_REPO_PASSWORD environment "
        "variable",
    )

    return parser



# pylint: disable=E1101,W0603
# Standard Library
import json
import os

# Third Party
from celery import Celery
from celery.utils.log import get_task_logger
from resc_backend.constants import TEMP_RULE_DIR_FILE, TEMP_RULE_FILE, TEMP_RULE_REPO_FILE
from resc_backend.resc_web_service.schema.repository import Repository

# First Party
from vcs_scanner.common import initialise_logs, load_vcs_instances
from vcs_scanner.constants import LOG_FILE_PATH
from vcs_scanner.helpers.environment_wrapper import validate_environment
from vcs_scanner.helpers.providers.rule_file import RuleFileProvider
from vcs_scanner.helpers.providers.rule_tag import RuleTagProvider
from vcs_scanner.model import RepositoryRuntime
from vcs_scanner.output_modules.rws_api_writer import RESTAPIWriter
from vcs_scanner.secret_scanners.configuration import (
    GITLEAKS_PATH,
    RABBITMQ_DEFAULT_VHOST,
    RABBITMQ_PASSWORD,
    RABBITMQ_QUEUE,
    RABBITMQ_SERVICE_HOST,
    RABBITMQ_USERNAME,
    REQUIRED_ENV_VARS,
    RESC_API_NO_AUTH_SERVICE_HOST,
    RESC_API_NO_AUTH_SERVICE_PORT,
    RESC_IGNORE_TAGS,
    RESC_INCLUDE_TAGS,
    VCS_INSTANCES_FILE_PATH,
)
from vcs_scanner.secret_scanners.secret_scanner import SecretScanner

env_variables = validate_environment(REQUIRED_ENV_VARS)
app = Celery(
    "secret_scanner_worker",
    broker="amqp://"
    + f"{env_variables[RABBITMQ_USERNAME]}"
    + ":"
    + f"{env_variables[RABBITMQ_PASSWORD]}"
    + "@"
    + f"{env_variables[RABBITMQ_SERVICE_HOST]}"
    + "/"
    + f"{env_variables[RABBITMQ_DEFAULT_VHOST]}",
)
app.conf.update({"worker_hijack_root_logger": False})
app.conf.update({"broker_connection_retry": True})
app.conf.update({"broker_connection_max_retries": 100})

logger = get_task_logger(__name__)
logger_config = initialise_logs(LOG_FILE_PATH)
rabbitmq_queue = env_variables[RABBITMQ_QUEUE]
rws_url = f"http://{env_variables[RESC_API_NO_AUTH_SERVICE_HOST]}:{env_variables[RESC_API_NO_AUTH_SERVICE_PORT]}"
rws_writer: RESTAPIWriter = RESTAPIWriter(rws_url=rws_url)

VCS_INSTANCES_LIST = None
VCS_INSTANCES = None
DOWNLOADED_RULE_PACK_VERSION = None


@app.task(name="scan_repository", Queue=rabbitmq_queue)
def scan_repository(repository):
    global VCS_INSTANCES_LIST, VCS_INSTANCES, DOWNLOADED_RULE_PACK_VERSION
    if not VCS_INSTANCES_LIST:
        VCS_INSTANCES_LIST = load_vcs_instances(env_variables[VCS_INSTANCES_FILE_PATH])
    if not VCS_INSTANCES:
        VCS_INSTANCES = rws_writer.write_vcs_instances(VCS_INSTANCES_LIST)
    if not DOWNLOADED_RULE_PACK_VERSION:
        DOWNLOADED_RULE_PACK_VERSION = rws_writer.download_rule_pack()

    active_rule_pack_version = rws_writer.check_active_rule_pack_version(rule_pack_version=DOWNLOADED_RULE_PACK_VERSION)

    repository_runtime = RepositoryRuntime(**json.loads(repository))

    logger.info(
        f"Received repository to scan via the queue '{rabbitmq_queue}' => "
        f"{repository_runtime.project_key}/{repository_runtime.repository_name}"
    )
    try:
        vcs_instance = VCS_INSTANCES[repository_runtime.vcs_instance_name]

        repository = Repository(
            project_key=repository_runtime.project_key,
            repository_id=repository_runtime.repository_id,
            repository_name=repository_runtime.repository_name,
            repository_url=repository_runtime.repository_url,
            vcs_instance=vcs_instance.id_,
        )
        # Split the include_tags by comma if supplied
        include_tags = env_variables[RESC_INCLUDE_TAGS].split(",") if env_variables[RESC_INCLUDE_TAGS] else None

        # Split the ignore_tags by comma if supplied
        ignore_tags = env_variables[RESC_IGNORE_TAGS].split(",") if env_variables[RESC_IGNORE_TAGS] else None

        rule_tag_provider = RuleTagProvider()
        rule_tag_provider.load(TEMP_RULE_FILE)

        rest_api_writer = RESTAPIWriter(
            rws_url=rws_url, include_tags=include_tags, ignore_tags=ignore_tags, rule_tag_provider=rule_tag_provider
        )

        gitleaks_rules_provider = RuleFileProvider(TEMP_RULE_FILE)
        gitleaks_rules_provider.init(
            destination_rule_as_repo=TEMP_RULE_REPO_FILE,
            destination_rule_as_dir=TEMP_RULE_DIR_FILE,
        )

        secret_scanner = SecretScanner(
            gitleaks_binary_path=env_variables[GITLEAKS_PATH],
            gitleaks_rules_provider=gitleaks_rules_provider,
            rule_pack_version=active_rule_pack_version,
            output_plugin=rest_api_writer,
            repository=repository,
            username=vcs_instance.username,
            personal_access_token=vcs_instance.token,
            force_base_scan=os.getenv("FORCE_BASE_SCAN", "false").lower() in "true",
            latest_commit=repository_runtime.latest_commit,
        )

        secret_scanner.run_scan(as_dir=True, as_repo=True)
    except KeyError:
        logger.error(
            f"No configuration found for vcs instance {repository_runtime.vcs_instance_name}, "
            f"unable to scan {repository_runtime.project_key}/{repository_runtime.repository_name}"
        )



# Standard Library
import getpass
import json
import logging.config
import os
from argparse import ArgumentParser, Namespace
from urllib.parse import urlparse

# Third Party
from resc_backend.resc_web_service.schema.vcs_provider import VCSProviders

# First Party
from vcs_scanner.common import get_rule_pack_version_from_file, initialise_logs
from vcs_scanner.constants import (
    CLI_VCS_AZURE,
    CLI_VCS_BITBUCKET,
    CLI_VCS_LOCAL_SCAN,
    LOG_FILE_PATH_CLI,
)
from vcs_scanner.helpers.cli import create_cli_argparser
from vcs_scanner.helpers.providers.rule_file import RuleFileProvider
from vcs_scanner.model import RepositoryRuntime
from vcs_scanner.output_modules.rws_api_writer import RESTAPIWriter
from vcs_scanner.output_modules.stdout_writer import STDOUTWriter
from vcs_scanner.secret_scanners.git_operation import read_repo_from_local
from vcs_scanner.secret_scanners.secret_scanner import SecretScanner

logger_config = initialise_logs(LOG_FILE_PATH_CLI)
logger = logging.getLogger(__name__)

FAKE_COMMIT = "hash"
FAKE_URL = "http://fake-host.none"


def deserialize_repository_from_file(filepath: str) -> RepositoryRuntime:
    with open(filepath, encoding="utf-8") as repo_file:
        repository_str: str = repo_file.read()
    repository: RepositoryRuntime = RepositoryRuntime(**json.loads(repository_str))
    return repository


def get_repository_name_from_url(repo_url: str) -> str:
    """
        Get repository name from given URL, taking the last segment of the url as name
    :param repo_url:
        Full url to the repository
    :return: str.
        The output will the name of the repository based on the url
    """
    url = urlparse(repo_url)
    if url.path.split("/")[-1] == "":
        return url.path.split("/")[-2]
    return url.path.split("/")[-1]


def validate_cli_arguments(args: Namespace):  # pylint: disable=R0912
    """
        Validate the CLI arguments given
    :param args:
        Namespace object containing the arguments parsed from the CLI
    :return: args or False.
        The output will be the args given, unless validation fails then it contains False
    """
    valid_arguments = True
    # Prompt for the password for a remote repo if username is specified
    if args.command == "repo" and args.repository_location == "remote" and args.username:
        if "RESC_REPO_PASSWORD" in os.environ:
            args.password = os.environ["RESC_REPO_PASSWORD"]
        else:
            args.password = getpass.getpass("Password:")

    # Derive the repository name from the directory or url if not provided
    if args.command == "repo" and args.repository_location == "remote" and not args.repo_name:
        args.repo_name = get_repository_name_from_url(args.repo_url)
    elif args.command == "dir" or (
        args.command == "repo" and args.repository_location == "local" and not args.repo_name
    ):
        if not os.path.isdir(args.dir.absolute()):
            logger.error(f"The directory {args.dir.absolute()} does not exist")
            valid_arguments = False
        args.repo_name = os.path.split(args.dir.absolute())[1]

    # Split the include_tags by comma if supplied
    args.include_tags = args.include_tags.split(",") if args.include_tags else None

    # Split the ignore_tags by comma if supplied
    args.ignore_tags = args.ignore_tags.split(",") if args.ignore_tags else None

    if not valid_arguments:
        return False

    return args


def scan_repository_from_cli():
    """
    Startup command for the CLI, parsing arguments and starting the process
    """
    parser: ArgumentParser = create_cli_argparser()
    args: Namespace = parser.parse_args()
    args = validate_cli_arguments(args)

    if args.verbose:
        logger_config.setLevel(logging.DEBUG)
    else:
        logger_config.setLevel(logging.INFO)

    if args.command == "dir":
        logger.info(f"Scanning directory {args.dir.absolute()}")
        scan_directory(args)
    elif args.command == "repo":
        if args.repository_location == "local":
            logger.info(f"Scanning repository local {args.dir.absolute()}")
            args.repo_url = fetch_url_from_dot_git_config(args.dir.absolute())
            args.username = None
            args.password = None
        elif args.repository_location == "remote":
            logger.info(f"Scanning repository remote {args.repo_url}")
        scan_repository(args)


def fetch_url_from_dot_git_config(path: str):
    if not os.path.exists(path / ".git/config"):
        return FAKE_URL

    return read_repo_from_local(path)


# TODO refactor and merge scan_directory / scan_repository to avoid code duplication.
def scan_directory(args: Namespace):
    """
        Start the process of scanning a non-git directory
    :param args:
        Namespace object containing the CLI arguments
    """
    repository = RepositoryRuntime(
        repository_url=FAKE_URL,
        repository_name="local",
        repository_id="local",
        project_key="local",
        vcs_instance_name="vcs_instance_name",
        latest_commit=FAKE_COMMIT,
    )

    output_plugin = STDOUTWriter.make(args)
    rule_pack_version = _get_rule_pack_version(args)

    if not rule_pack_version:
        rule_pack_version = "0.0.0"

    gitleaks_rules_provider = RuleFileProvider(args.gitleaks_rules_path, init=True)

    secret_scanner = SecretScanner(
        gitleaks_binary_path=args.gitleaks_path,
        gitleaks_rules_provider=gitleaks_rules_provider,
        rule_pack_version=rule_pack_version,
        output_plugin=output_plugin,
        repository=repository.convert_to_repository(vcs_instance_id=1),
        username="",
        personal_access_token="",
        local_path=f"{args.dir.absolute()}",
        # we force a base scan because it does not matter
        # in this use case: we are not sending data to RESC.
        force_base_scan=True,
    )

    secret_scanner.run_scan(as_dir=True)


def scan_repository(args: Namespace):
    """
        Start the process of scanning a git repository (remote or local)
    :param args:
        Namespace object containing the CLI arguments
    """
    vcs_type = guess_vcs_provider(args.repo_url)
    vcs_name = determine_vcs_name(args.repo_url, vcs_type)

    repository = RepositoryRuntime(
        repository_url=args.repo_url,
        repository_name=args.repo_name,
        repository_id=args.repo_name,
        project_key=args.repo_name,
        vcs_instance_name=vcs_name,
        latest_commit=FAKE_COMMIT,
    )

    if args.rws_url:
        output_plugin = RESTAPIWriter.make(args)
        rule_pack_version = output_plugin.download_rule_pack()

    else:
        output_plugin = STDOUTWriter.make(args)
        rule_pack_version = _get_rule_pack_version(args)

    if not rule_pack_version:
        rule_pack_version = "0.0.0"

    gitleaks_rules_provider = RuleFileProvider(args.gitleaks_rules_path, init=True)

    secret_scanner = SecretScanner(
        gitleaks_binary_path=args.gitleaks_path,
        gitleaks_rules_provider=gitleaks_rules_provider,
        rule_pack_version=rule_pack_version,
        output_plugin=output_plugin,
        repository=repository.convert_to_repository(vcs_instance_id=1),
        username=args.username,
        personal_access_token=args.password,
        local_path=f"{args.dir.absolute()}",
        force_base_scan=args.force_base_scan,
        latest_commit="unknown",
    )

    secret_scanner.run_scan(as_repo=True)


def guess_vcs_provider(repo_url: str) -> VCSProviders:
    """
        Guess the vcs provider based on the url given, defaulted to bitbucket
    :param repo_url:
        Full url of the repository
    :return: VCSProviders.
        The output will contain the guessed VCSProviders enum value
    """
    url = urlparse(repo_url)
    if "bitbucket" in url.netloc:
        return VCSProviders.BITBUCKET
    if "dev.azure" in url.netloc:
        return VCSProviders.AZURE_DEVOPS
    logger.warning("Unable to guess VCS_Provider, assuming it is bitbucket.")
    return VCSProviders.BITBUCKET


def determine_vcs_name(repo_url: str, vcs_type: VCSProviders) -> str:
    """
        Determine the vcs provider name based on the vcs_type given, defaulted to CLI_VCS_LOCAL_SCAN
    :param repo_url:
        Full url of the repository
    :param vcs_type:
        VCSProviders type of the repository
    :return: str.
        The output will contain the name of the vcs provider
    """
    vcs_name = CLI_VCS_LOCAL_SCAN
    if repo_url and repo_url is not FAKE_URL:
        if vcs_type == VCSProviders.AZURE_DEVOPS:
            vcs_name = CLI_VCS_AZURE
        elif vcs_type == VCSProviders.BITBUCKET:
            vcs_name = CLI_VCS_BITBUCKET
    return vcs_name


def _get_rule_pack_version(args: Namespace) -> str | None:
    with open(args.gitleaks_rules_path, encoding="utf-8") as rule_pack:
        return get_rule_pack_version_from_file(rule_pack.read())



from typing import Required, TypedDict


# Move me to Backend later
# https://github.com/gitleaks/gitleaks?tab=readme-ov-file#configuration
class Allowlist(TypedDict, total=False):
    description: Required[str]
    paths: list[str]
    regexes: list[str]


class RulesAllowlist(TypedDict, total=False):
    description: Required[str]
    regexTarget: str
    regexes: list[str]
    paths: list[str]
    stopwords: list[str]


class RuleToml(TypedDict, total=False):
    # from GitLeeks
    id: Required[str]
    description: Required[str]
    regex: Required[str]
    keywords: list[str]
    entropy: float
    path: str
    secretGroup: int
    allowlist: RulesAllowlist
    tags: Required[list[str]]
    # from RESC
    comment: str


class GitLeaksConfigToml(TypedDict):  # Implicit total = True
    title: str
    version: str
    allowlist: Allowlist
    rules: list[RuleToml]



import tomlkit


class RuleTagProvider:
    def __init__(self):
        self.toml_rule_file_path: str | None = None
        self.rule_tags: dict[str, list[str]] = {}

    def load(self, toml_rule_file_path: str) -> None:
        """
        Load a rule file into the provider.

        Args:
            toml_rule_file_path (str): path to the rule file.
        """
        self.toml_rule_file_path = toml_rule_file_path
        self.rule_tags = {}

    def get_rule_tags(self) -> dict[str, list[str]]:
        """
            Get the tags per rule from the .toml rule file, from self.toml_rule_file_path
        :return: dict.
            The output will contain a dictionary with the rule id as the key and the tags as a list in the value
        """
        if self.toml_rule_file_path is None:
            return {}

        if not self.rule_tags == {}:
            return self.rule_tags

        self.rule_tags = {}
        # read toml
        with open(self.toml_rule_file_path, encoding="utf-8") as toml_rule_file:
            toml_rule_dictionary = tomlkit.loads(toml_rule_file.read())
            # convert to dict
            for toml_rule in toml_rule_dictionary["rules"]:
                rule_id = toml_rule.get("id", None)
                if rule_id:
                    self.rule_tags[rule_id] = toml_rule.get("tags", [])
        return self.rule_tags



# Standard Library
import logging
import sys
from argparse import Namespace
from datetime import UTC, datetime

# Third Party
from prettytable import PrettyTable
from resc_backend.resc_web_service.schema.finding import FindingCreate
from resc_backend.resc_web_service.schema.repository import Repository
from resc_backend.resc_web_service.schema.scan import ScanRead
from resc_backend.resc_web_service.schema.scan_type import ScanType
from resc_backend.resc_web_service.schema.vcs_instance import VCSInstanceRead
from termcolor import colored

# First Party
from vcs_scanner.helpers.finding_action import FindingAction
from vcs_scanner.helpers.finding_filter import should_process_finding
from vcs_scanner.helpers.providers.ignore_list import IgnoredListProvider
from vcs_scanner.helpers.providers.rule_comment import RuleCommentProvider
from vcs_scanner.helpers.providers.rule_tag import RuleTagProvider
from vcs_scanner.model import VCSInstanceRuntime
from vcs_scanner.output_modules.output_module import OutputModule

logger = logging.getLogger(__name__)


class STDOUTWriter(OutputModule):
    def __init__(
        self,
        exit_code_warn: int,
        exit_code_block: int,
        include_tags: list[str] = None,
        ignore_tags: list[str] = None,
        working_dir: str = "",
        ignore_findings_providers: IgnoredListProvider = IgnoredListProvider(None),
        rule_tag_provider: RuleTagProvider = RuleTagProvider(),
        rule_comment_provider: RuleCommentProvider = RuleCommentProvider(),
    ):
        self.exit_code_warn: int = exit_code_warn
        self.exit_code_block: int = exit_code_block
        self.include_tags: list[str] = include_tags
        self.ignore_tags: list[str] = ignore_tags
        self.exit_code_success = 0
        self.working_dir = working_dir
        self.ignore_findings_providers: IgnoredListProvider = ignore_findings_providers
        self.rule_tag_provider: RuleTagProvider = rule_tag_provider
        self.rule_comment_provider: RuleCommentProvider = rule_comment_provider

    def load_rules(self, toml_rule_file_path: str) -> None:
        self.rule_tag_provider.load(toml_rule_file_path)
        self.rule_comment_provider.load(toml_rule_file_path)

    def write_vcs_instance(self, vcs_instance_runtime: VCSInstanceRuntime) -> VCSInstanceRead | None:
        vcs_instance = VCSInstanceRead(
            id_=1,
            name=vcs_instance_runtime.name,
            provider_type=vcs_instance_runtime.provider_type,
            hostname=vcs_instance_runtime.hostname,
            port=vcs_instance_runtime.port,
            scheme=vcs_instance_runtime.scheme,
            exceptions=vcs_instance_runtime.exceptions,
            scope=vcs_instance_runtime.scope,
            organization=vcs_instance_runtime.organization,
        )
        logger.info(f"Scanning vcs instance {vcs_instance.name}")
        return vcs_instance

    def write_repository(self, repository: Repository) -> Repository:
        logger.info(f"Scanning repository {repository.project_key}/{repository.repository_name}")
        return repository

    @staticmethod
    def _determine_finding_action(finding: FindingCreate, rule_tags: dict = None) -> FindingAction:
        """
            Determine the action to take for the finding, based on the rule tags
        :param finding:
            FindingCreate instance of the finding
        :param rule_tags:
            Dictionary containing all the rules and there respective tags
        :return: FindingAction.
            FindingAction to take for this finding
        """
        rule_action = FindingAction.INFO
        if rule_tags and FindingAction.WARN in rule_tags.get(finding.rule_name, []):
            rule_action = FindingAction.WARN
        if rule_tags and FindingAction.BLOCK in rule_tags.get(finding.rule_name, []):
            rule_action = FindingAction.BLOCK

        return rule_action

    @staticmethod
    def _determine_if_ignored(
        rule_action: FindingAction,
        finding: FindingCreate,
        ignore_dictionary: dict,
        working_dir: str,
    ) -> FindingAction:
        """
            Determine whether to ignore a blocker or not.
        :param rule_action:
            FindingAction containing the decision depending of rules.
        :param finding:
            FindingCreate instance of the finding
        :param ignore_dictionary:
            Dictionary containing all the list of ignored blockers
        :return: FindingAction.
            FindingAction to take for this finding
        """
        if rule_action is not FindingAction.BLOCK:
            logger.info(f"{rule_action}: {finding.file_path}")
            return rule_action

        if working_dir is None or working_dir == "":
            working_dir = "/"

        working_dir = str(working_dir)
        if working_dir[-1] != "/":
            working_dir = working_dir + "/"

        path_length: int = len(working_dir)
        finding_path: str = str(finding.file_path)
        if finding_path[:path_length] == working_dir:
            finding_path = finding_path[path_length:]

        key: str = finding_path + "|" + finding.rule_name + "|" + str(finding.line_number)
        if key in ignore_dictionary:
            return FindingAction.IGNORED

        # Wildcard check.
        key: str = finding_path + "|" + finding.rule_name + "|*"
        if key in ignore_dictionary:
            return FindingAction.IGNORED

        return rule_action

    def write_findings(
        self, scan_id: int, repository_id: int, scan_findings: list[FindingCreate], repository_name: str = ""
    ):
        """
            Write the findings to the STDOUT in a nice table and set the exit code based on the FindingActions found
        :param scan_id:
            id of the scan in question
        :param repository_id:
            id of the repository in question
        :param scan_findings:
            List of FindingCreate of all the findings from the scan
        """
        # Initialize table
        output_table = PrettyTable()
        output_table.field_names = [
            "Level",
            "Rule",
            "Line",
            "Position",
            "File path",
            "Comment",
        ]
        output_table.align = "l"
        output_table.align["Line"] = "r"

        block_count = 0
        warn_count = 0
        info_count = 0

        exit_code = self.exit_code_success
        rule_tags = self.rule_tag_provider.get_rule_tags()
        ignore_dictionary = self.ignore_findings_providers.get_ignore_list()
        for finding in scan_findings:
            logger.debug(finding.commit_id)
            should_process = should_process_finding(
                finding=finding,
                rule_tags=rule_tags,
                include_tags=self.include_tags,
                ignore_tags=self.ignore_tags,
            )
            if should_process:
                finding_action = self._determine_finding_action(finding, rule_tags)
                finding_action = self._determine_if_ignored(
                    finding_action, finding, ignore_dictionary, self.working_dir
                )

                if finding_action == FindingAction.BLOCK:
                    finding_action_value = colored(finding_action.value, "red", attrs=["bold"])
                    block_count += 1
                elif finding_action in [FindingAction.WARN, FindingAction.IGNORED]:
                    finding_action_value = colored(finding_action.value, "light_red", attrs=["bold"])
                    warn_count += 1
                elif finding_action == FindingAction.INFO:
                    finding_action_value = colored(finding_action.value, "light_yellow", attrs=["bold"])
                    info_count += 1
                else:
                    finding_action_value = finding_action.value
                    info_count += 1

                if exit_code != self.exit_code_block:
                    if exit_code == self.exit_code_success and finding_action in [
                        FindingAction.WARN,
                        FindingAction.IGNORED,
                    ]:
                        exit_code = self.exit_code_warn
                    elif finding_action == FindingAction.BLOCK:
                        exit_code = self.exit_code_block

                comment = self.rule_comment_provider.get_comment().get(finding.rule_name, "")

                output_table.add_row(
                    [
                        finding_action_value,
                        finding.rule_name,
                        finding.line_number,
                        f"{finding.column_start}-{finding.column_end}",
                        finding.file_path,
                        comment,
                    ]
                )

        logger.info(f"\n{output_table.get_string(sortby='Level')}")

        logger.info(
            f"Findings detected : Total - {block_count+warn_count+info_count}, Block - {block_count}, "
            f"Warn - {warn_count}, Info - {info_count}"
        )

        if exit_code == self.exit_code_success:
            logger.info(f"Findings threshold check results: {colored('PASS', 'light_green', attrs=['bold'])}")
        elif exit_code == self.exit_code_block:
            logger.info(f"Findings threshold check results: {colored('FAIL', 'red', attrs=['bold'])}")
            logger.info(
                colored(
                    f"Scan failed due to policy violations: [Block:{block_count}]",
                    "red",
                    attrs=["bold"],
                )
            )
        elif exit_code == self.exit_code_warn:
            logger.info(f"Findings threshold check results: {colored('PASS', 'light_green', attrs=['bold'])}")
            logger.info(
                colored(
                    f"Warning for policy violations: [Warn:{warn_count}]",
                    "light_red",
                    attrs=["bold"],
                )
            )

        sys.exit(exit_code)

    def write_scan(
        self,
        scan_type_to_run: ScanType,
        last_scanned_commit: str,
        scan_timestamp: datetime,
        repository: Repository,
        rule_pack: str,
    ) -> ScanRead | None:
        logger.info(f"Running {scan_type_to_run} scan on repository {repository.repository_url}")
        return ScanRead(
            last_scanned_commit="NONE",
            timestamp=datetime.now(UTC),
            repository_id=1,
            id_=1,
            rule_pack=rule_pack,
        )

    def get_last_scan_for_repository(self, repository: Repository) -> ScanRead | None:
        return None

    @staticmethod
    def make(args: Namespace) -> "STDOUTWriter":
        """
            Get the STDOUT writer given the args provided.

        :param args:
            Namespace object containing the CLI arguments
        """
        rule_tag_provider = RuleTagProvider()
        rule_tag_provider.load(args.gitleaks_rules_path)

        rule_comment_provider = RuleCommentProvider()
        rule_comment_provider.load(args.gitleaks_rules_path)

        ignored_finding_provider = IgnoredListProvider(args.ignored_blocker_path)

        output_plugin = STDOUTWriter(
            exit_code_warn=args.exit_code_warn,
            exit_code_block=args.exit_code_block,
            include_tags=args.include_tags,
            ignore_tags=args.ignore_tags,
            working_dir=args.dir,
            ignore_findings_providers=ignored_finding_provider,
            rule_tag_provider=rule_tag_provider,
            rule_comment_provider=rule_comment_provider,
        )

        return output_plugin
